import os
from argparse import ArgumentParser
from glob import glob
from os.path import basename, join

import torch
import torchaudio
from tqdm import tqdm

from auv.model import AUV


parser = ArgumentParser()
parser.add_argument("--input-dir", type=str, default=".")
parser.add_argument("--output-dir", type=str, default="outputs")
parser.add_argument("--ckpt", type=str, default="auv.pt")
parser.add_argument("--device", type=str, default="cuda:0")
parser.add_argument("--bf16", action="store_true")
args = parser.parse_args()

input_dir = args.input_dir
output_dir = args.output_dir
ckpt = args.ckpt
device = args.device
bf16 = args.bf16


def main():
    print(f"Load codec ckpt from {ckpt}")
    print(f"Use bf16: {bf16}")
    auv = AUV()
    auv.from_pretrained(ckpt)
    auv = auv.to(device)
    auv.eval()
    target_sr = auv.tokenizer.sample_rate

    wav_paths = glob(join(input_dir, "*.wav"))

    os.makedirs(output_dir, exist_ok=True)
    for wav_path in tqdm(wav_paths):
        target_wav_path = join(output_dir, basename(wav_path))
        wav, sr = torchaudio.load(wav_path)
        wav = wav.to(device)
        if wav.shape[0] > 1:
            wav = torch.mean(wav, dim=0, keepdim=True)

        data = {
            "sample": wav,
            "sample_rate": sr,
        }

        with torch.autocast(
            device_type="cuda",
            dtype=torch.bfloat16,
            enabled=bf16,
        ):
            enc_res = auv.encode(data)
            vq_post_emb = enc_res["quantized"]
            # vq_code = enc_res["tokens"]
            # vq_pre_emb = enc_res["before_quantize"]

            recon = auv.decode(vq_post_emb)
            # recon = auv.decode_tokens(vq_code.transpose(1, 2))

        torchaudio.save(target_wav_path, recon[0].cpu(), target_sr)


if __name__ == "__main__":
    main()



import torch
from torch import nn


class STFT(nn.Module):
    def __init__(
        self,
        n_fft: int,
        hop_length: int,
        win_length: int,
        center=True,
    ):
        super().__init__()
        self.center = center
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        window = torch.hann_window(win_length)
        self.register_buffer("window", window)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (B, T * hop_length)

        if not self.center:
            pad = self.win_length - self.hop_length
            x = nn.functional.pad(x, (pad // 2, pad // 2), mode="reflect")

        stft_spec = torch.stft(
            x,
            self.n_fft,
            hop_length=self.hop_length,
            win_length=self.win_length,
            window=self.window,
            center=self.center,
            onesided=True,
            return_complex=True,
        )
        stft_spec = torch.view_as_real(stft_spec)  # (B, n_fft // 2 + 1, T, 2)

        rea = stft_spec[:, :, :, 0]  # (B, n_fft // 2 + 1, T, 2)
        imag = stft_spec[:, :, :, 1]  # (B, n_fft // 2 + 1, T, 2)

        log_mag = torch.log(
            torch.abs(torch.sqrt(torch.pow(rea, 2) + torch.pow(imag, 2))) + 1e-5
        )  # (B, n_fft // 2 + 1, T)
        phase = torch.atan2(imag, rea)  # (B, n_fft // 2 + 1, T)

        return log_mag, phase


class STFTHead(nn.Module):
    def __init__(self, dim: int, n_fft: int, hop_length: int, **kwargs):
        super().__init__()
        self.stft = STFT(n_fft=n_fft, hop_length=hop_length, win_length=n_fft, center=False)
        inp_dim = n_fft + 2
        self.inp = nn.Linear(inp_dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        log_mag, phase = self.stft(x)
        x = torch.cat((log_mag, phase), dim=1).transpose(1, 2)
        x = self.inp(x)  # (B, T, C)
        return x


class ISTFT(nn.Module):
    # https://github.com/pytorch/pytorch/issues/62323
    # fix padding type to "same" here

    def __init__(self, n_fft: int, hop_length: int, win_length: int):
        super().__init__()
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        window = torch.hann_window(win_length)
        self.register_buffer("window", window)

    def forward(self, spec: torch.Tensor) -> torch.Tensor:
        pad = (self.win_length - self.hop_length) // 2

        assert spec.dim() == 3, "Expected a 3D tensor as input"
        B, N, T = spec.shape

        # Inverse FFT
        ifft = torch.fft.irfft(spec, self.n_fft, dim=1, norm="backward")
        ifft = ifft * self.window[None, :, None]

        # Overlap and Add
        output_size = (T - 1) * self.hop_length + self.win_length
        y = nn.functional.fold(
            ifft,
            output_size=(1, output_size),
            kernel_size=(1, self.win_length),
            stride=(1, self.hop_length),
        )[:, 0, 0, pad:-pad]

        # Window envelope
        window_sq = self.window.square().expand(1, T, -1).transpose(1, 2)
        window_envelope = nn.functional.fold(
            window_sq,
            output_size=(1, output_size),
            kernel_size=(1, self.win_length),
            stride=(1, self.hop_length),
        ).squeeze()[pad:-pad]

        # Normalize
        assert (window_envelope > 1e-11).all()
        y = y / window_envelope

        return y


class ISTFTHead(nn.Module):
    def __init__(self, dim: int, n_fft: int, hop_length: int, **kwargs):
        super().__init__()
        out_dim = n_fft + 2
        self.out = nn.Linear(dim, out_dim)
        self.istft = ISTFT(n_fft=n_fft, hop_length=hop_length, win_length=n_fft)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.out(x).transpose(1, 2)
        mag, p = x.chunk(2, dim=1)
        mag = torch.exp(mag)
        mag = torch.clip(mag, max=1e2)
        x = torch.cos(p)
        y = torch.sin(p)
        S = mag * (x + 1j * y)
        audio = self.istft(S)
        return audio



import torch
from torch import nn

from auv.modules.factorized_vector_quantize import FactorizedVectorQuantize


REGISTERED_VQ_TYPE = {
    "fvq": FactorizedVectorQuantize,
}


class ResidualVQ(nn.Module):
    def __init__(
        self,
        *,
        vq_type,
        num_quantizers,
        codebook_size,
        **kwargs,
    ):
        super().__init__()
        try:
            VQ = REGISTERED_VQ_TYPE[vq_type]
        except KeyError:
            raise NotImplementedError(f"{vq_type} is not implemented.")

        if isinstance(codebook_size, int):
            codebook_size = [codebook_size] * num_quantizers
        self.quantizers = nn.ModuleList([VQ(codebook_size=size, **kwargs) for size in codebook_size])

        self.num_quantizers = num_quantizers

    def forward(self, x, codebook_partitions=None):
        quantized_out = 0.0
        residual = x

        all_losses = []
        all_indices = []

        for idx, quantizer in enumerate(self.quantizers):
            z_q_i, indices_i, vq_loss_i = quantizer(
                residual,
                codebook_partitions=codebook_partitions,
            )
            quantized_out = quantized_out + z_q_i
            residual = residual - z_q_i

            vq_loss_i = vq_loss_i.mean()

            all_indices.append(indices_i)
            all_losses.append(vq_loss_i)

        all_losses, all_indices = map(torch.stack, (all_losses, all_indices))
        return quantized_out, all_indices, all_losses

    def vq2emb(self, vq, proj=True):
        # [B, T, num_quantizers]
        quantized_out = 0.0
        for idx, quantizer in enumerate(self.quantizers):
            z_q_i = quantizer.vq2emb(vq[:, :, idx], proj=proj)
            quantized_out = quantized_out + z_q_i
        return quantized_out

    def get_emb(self):
        embs = []
        for idx, quantizer in enumerate(self.quantizers):
            embs.append(quantizer.get_emb())
        return embs



import math
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class RotaryEmbedding(nn.Module):
    def __init__(self, dim: int, max_position_embeddings: int = 2048, base: int = 10000, device=None):
        super().__init__()
        self.dim = dim
        self.max_position_embeddings = max_position_embeddings
        self.base = base
        inv_freq = 1.0 / (self.base ** (torch.arange(0, self.dim, 2).float().to(device) / self.dim))
        self.register_buffer("inv_freq", inv_freq, persistent=False)
        # Build here to make `torch.jit.trace` work.
        self._set_cos_sin_cache(
            seq_len=max_position_embeddings, device=self.inv_freq.device, dtype=torch.get_default_dtype()
        )

    @torch.amp.autocast("cuda", enabled=False)
    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len
        t = torch.arange(self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype)
        freqs = torch.einsum("i,j->ij", t, self.inv_freq)
        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer("cos_cached", emb.cos()[None, None, :, :].to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin()[None, None, :, :].to(dtype), persistent=False)

    def forward(self, x, seq_len=None):
        if seq_len is None:
            seq_len = x.shape[-2]
        if seq_len > self.max_seq_len_cached:
            self._set_cos_sin_cache(seq_len=seq_len, device=x.device, dtype=x.dtype)
        return (
            self.cos_cached[:, :, :seq_len, ...].to(dtype=x.dtype),
            self.sin_cached[:, :, :seq_len, ...].to(dtype=x.dtype),
        )

    def rotate_half(self, x):
        """Rotates half the hidden dims of the input."""
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return torch.cat((-x2, x1), dim=-1)

    @torch.amp.autocast("cuda", enabled=False)
    def apply_rotary_pos_emb(self, qk, cos, sin, position_ids):
        cos = cos.squeeze(1).squeeze(0)  # [seq_len, dim]
        sin = sin.squeeze(1).squeeze(0)  # [seq_len, dim]
        cos = cos[position_ids].unsqueeze(1)  # [bs, 1, seq_len, dim]
        sin = sin[position_ids].unsqueeze(1)  # [bs, 1, seq_len, dim]
        qk_embed = (qk * cos) + (self.rotate_half(qk) * sin)
        return qk_embed


class MultiheadRMSNorm(nn.Module):
    def __init__(self, dim, heads):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(heads, 1, dim))

    def forward(self, x):
        return F.normalize(x, p=2, dim=-1) * self.gamma


class Attention(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        num_heads: int = 16,
        head_dim: int = 64,
        dropout: float = 0.0,
        qk_norm: bool = False,
        qk_norm_scale: float = 10.0,
        qkv_bias: bool = False,
        max_position_embeddings: int = 2048,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = head_dim
        dim_inner = num_heads * head_dim
        self.qk_norm = qk_norm
        if qk_norm:
            self.q_norm = MultiheadRMSNorm(head_dim, heads=num_heads)
            self.k_norm = MultiheadRMSNorm(head_dim, heads=num_heads)
        self.scale = qk_norm_scale if qk_norm else 1.0 / math.sqrt(head_dim)
        self.dropout = dropout

        self.to_qkv = nn.Linear(hidden_size, 3 * dim_inner, bias=qkv_bias)
        self.to_out = nn.Linear(dim_inner, hidden_size)
        self.attn_dropout = nn.Dropout(dropout)
        self.rotary_embed = RotaryEmbedding(self.head_dim, max_position_embeddings=max_position_embeddings)

    def forward(
        self,
        hidden_states: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[torch.Tensor]] = None,
        use_cache: bool = False,
    ):
        batch_size, q_len, _ = hidden_states.size()
        q, k, v = self.to_qkv(hidden_states).chunk(3, dim=-1)
        q, k, v = map(lambda x: rearrange(x, "b n (h d) -> b h n d", h=self.num_heads), (q, k, v))
        if self.qk_norm:
            q = self.q_norm(q)
            k = self.k_norm(k)
        kv_seq_len = k.shape[-2]
        if past_key_values is not None:
            kv_seq_len += past_key_values[0].shape[-2]

        # position embeddings
        cos, sin = self.rotary_embed(k, seq_len=kv_seq_len)
        if position_ids is None:
            position_ids = torch.arange(q.shape[-2], dtype=torch.long, device=q.device).unsqueeze(0)
        q, k = map(lambda x: self.rotary_embed.apply_rotary_pos_emb(x, cos, sin, position_ids), (q, k))

        # deal history
        if past_key_values is not None:
            k = torch.cat([past_key_values[0], k], dim=2)
            v = torch.cat([past_key_values[1], v], dim=2)
        past_key_values = (k, v) if use_cache else None

        # attention mask
        if mask is not None:  # 0 for padding position
            if mask.size() != (batch_size, q_len, kv_seq_len):
                raise ValueError(
                    f"Attention mask should be of size {(batch_size, q_len, kv_seq_len)}, got {mask.size()}"
                )
            mask = mask.unsqueeze(1).expand(-1, self.num_heads, -1, -1)

        attn_output = F.scaled_dot_product_attention(
            q,
            k,
            v,
            attn_mask=mask,
            dropout_p=self.dropout if self.training else 0.0,
            scale=self.scale,
        )

        attn_output = rearrange(attn_output, "b h n d -> b n (h d)")
        attn_output = self.to_out(attn_output)

        return attn_output, past_key_values


class FeedForward(nn.Module):
    def __init__(self, hidden_size: int, ffn_hidden_size: int):
        super().__init__()
        self.hidden_size = hidden_size
        self.ffn_hidden_size = ffn_hidden_size
        self.up_proj = nn.Linear(hidden_size, ffn_hidden_size)
        self.down_proj = nn.Linear(ffn_hidden_size, hidden_size)
        self.act_fn = nn.GELU(approximate="tanh")

    def forward(self, x):
        x = self.act_fn(self.up_proj(x))
        x = self.down_proj(x)
        return x


class ConformerConvolutionModule(nn.Module):
    def __init__(self, hidden_size, conv_kernel_size=31, causal=False):
        super().__init__()
        self.causal = causal
        self.conv_kernel_size = conv_kernel_size
        # norm before conv
        self.layer_norm = nn.LayerNorm(hidden_size)
        self.pointwise_conv1 = nn.Conv1d(hidden_size, 2 * hidden_size, kernel_size=1, stride=1, padding=0, bias=False)
        self.glu = nn.GLU(dim=1)
        # depth-wise conv
        if self.causal:
            padding = 0
            self.lorder = self.conv_kernel_size - 1
        else:
            assert (self.conv_kernel_size % 2) == 1, "conformer conv_kernel_size should be odd for SAME padding"
            padding = (self.conv_kernel_size - 1) // 2
            self.lorder = 0
        self.depthwise_conv = nn.Conv1d(
            hidden_size,
            hidden_size,
            kernel_size=self.conv_kernel_size,
            stride=1,
            padding=padding,
            groups=hidden_size,
            bias=False,
        )
        self.conv_norm = nn.BatchNorm1d(hidden_size)
        self.activation = nn.GELU(approximate="tanh")
        # pointwise conv 2
        self.pointwise_conv2 = nn.Conv1d(hidden_size, hidden_size, kernel_size=1, stride=1, padding=0, bias=False)

    def forward(self, hidden_states, mask: Optional[torch.Tensor] = None, cache: Optional[torch.Tensor] = None):
        """
        hidden_states: [B, T, F]
        mask: [B, T]
        cache: [B, F, h], for depthwise-conv
        """

        hidden_states = self.layer_norm(hidden_states)
        hidden_states = hidden_states.transpose(1, 2)

        # pointwise-conv 1
        hidden_states = self.pointwise_conv1(hidden_states)
        hidden_states = self.glu(hidden_states)

        # mask pad & cache
        if mask is not None:
            mask = mask.unsqueeze(1)
            hidden_states = hidden_states.masked_fill(~mask, 0.0)
        prepend_len = 0
        if cache is not None:
            prepend_len = cache.shape[2]
            hidden_states = torch.cat([cache, hidden_states], dim=2)
        cache_len = self.lorder if self.lorder > 0 else (self.conv_kernel_size - 1) // 2
        new_cache = hidden_states[:, :, -cache_len:]

        # depthwise conv
        if self.lorder > 0:
            hidden_states = F.pad(hidden_states, (self.lorder, 0), "constant", 0.0)
        hidden_states = self.depthwise_conv(hidden_states)
        hidden_states = hidden_states[:, :, prepend_len:]

        # batch norm & activation
        hidden_states = self.conv_norm(hidden_states)
        hidden_states = self.activation(hidden_states)

        # pointwise conv 2
        hidden_states = self.pointwise_conv2(hidden_states)
        if mask is not None:
            hidden_states = hidden_states.masked_fill(~mask, 0.0)
        hidden_states = hidden_states.transpose(1, 2)
        return hidden_states, new_cache


class ConformerEncoderLayer(nn.Module):
    def __init__(
        self,
        hidden_size=1024,
        ffn_mult=4,
        head_dim=64,
        num_heads=None,
        qk_norm=False,
        qk_norm_scale=10.0,
        conv_kernel_size=31,
        max_position_embeddings=4096,
        causal=False,
    ):
        super().__init__()
        embed_dim = hidden_size
        if num_heads is None:
            num_heads = embed_dim // head_dim

        # ffn 1
        self.ffn1_layer_norm = nn.LayerNorm(embed_dim)
        self.ffn1 = FeedForward(embed_dim, hidden_size * ffn_mult)

        # self attention
        self.attn_layer_norm = nn.LayerNorm(embed_dim)
        self.attn = Attention(
            embed_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            qk_norm=qk_norm,
            qk_norm_scale=qk_norm_scale,
            max_position_embeddings=max_position_embeddings,
        )

        # conformer convolution
        self.conv_module = ConformerConvolutionModule(
            hidden_size=hidden_size, conv_kernel_size=conv_kernel_size, causal=causal
        )

        # ffn 2
        self.ffn2_layer_norm = nn.LayerNorm(embed_dim)
        self.ffn2 = FeedForward(embed_dim, hidden_size * ffn_mult)
        self.final_layer_norm = nn.LayerNorm(embed_dim)

    def forward(
        self,
        hidden_states,
        attention_mask: Optional[torch.Tensor] = None,
        mask_pad: Optional[torch.Tensor] = None,  # for conv module
        position_ids: Optional[torch.LongTensor] = None,
        caches: Optional[Tuple[torch.Tensor]] = None,
        use_cache: bool = False,
    ):
        if caches is not None:  # should be tuple of (past_keys, past_values, cnn_caches)
            past_key_values, cnn_caches = caches[0:2], caches[2]
        else:
            past_key_values, cnn_caches = None, None

        # ffn1
        residual = hidden_states
        hidden_states = self.ffn1_layer_norm(hidden_states)
        hidden_states = self.ffn1(hidden_states)
        hidden_states = residual + hidden_states * 0.5

        # self attn
        residual = hidden_states
        hidden_states = self.attn_layer_norm(hidden_states)
        hidden_states, past_key_values = self.attn(
            hidden_states,
            mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            use_cache=use_cache,
        )
        hidden_states = residual + hidden_states

        # convolution  module
        residual = hidden_states
        hidden_states, cnn_caches = self.conv_module(hidden_states, mask=mask_pad, cache=cnn_caches)
        hidden_states = residual + hidden_states

        #  ffn 2
        residual = hidden_states
        hidden_states = self.ffn2_layer_norm(hidden_states)
        hidden_states = residual + 0.5 * self.ffn2(hidden_states)

        # out
        hidden_states = self.final_layer_norm(hidden_states)
        if use_cache:
            new_cache = past_key_values + (cnn_caches,)
        else:
            new_cache = None
        return hidden_states, new_cache


class ConformerEncoder(nn.Module):
    def __init__(
        self,
        num_layers=12,
        hidden_size=1024,
        ffn_mult=4,
        head_dim=64,
        num_heads=None,
        qk_norm=False,
        qk_norm_scale=10.0,
        conv_kernel_size=31,
        max_position_embeddings=4096,
        causal=False,
        **kwargs,
    ):
        super().__init__()
        self.num_layers = num_layers
        self.hidden_size = hidden_size

        # layers
        self.layers = nn.ModuleList(
            [
                ConformerEncoderLayer(
                    hidden_size=hidden_size,
                    ffn_mult=ffn_mult,
                    head_dim=head_dim,
                    num_heads=num_heads,
                    qk_norm=qk_norm,
                    qk_norm_scale=qk_norm_scale,
                    conv_kernel_size=conv_kernel_size,
                    max_position_embeddings=max_position_embeddings,
                    causal=causal,
                )
                for _ in range(num_layers)
            ]
        )
        self.layer_norm = nn.LayerNorm(hidden_size)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        mask_pad: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        caches: Optional[List[Tuple[torch.Tensor]]] = None,
        output_hidden_states: bool = False,
        use_cache: bool = False,
    ):
        all_hidden_states = () if output_hidden_states else None
        new_caches = () if use_cache else None
        batch_size, seq_len, _, device = *hidden_states.shape, hidden_states.device
        seq_len_with_past, past_len = seq_len, 0
        if caches is not None:
            past_len = caches[0][0].shape[-2]
            seq_len_with_past += past_len

        # position id
        if position_ids is None:
            position_ids = torch.arange(past_len, past_len + seq_len, dtype=torch.long, device=device)
            position_ids = position_ids.unsqueeze(0)
        else:
            position_ids = position_ids.view(-1, seq_len).long()

        # attention mask
        if attention_mask is not None:
            if attention_mask.size() != (batch_size, seq_len, seq_len):
                raise ValueError(
                    f"given attn mask should be of size {(batch_size, seq_len, seq_len)}, got {attention_mask.size()}"
                )
            if past_len > 0:
                extra = torch.ones(batch_size, seq_len, past_len, dtype=attention_mask.dtype, device=device)
                attention_mask = torch.cat([extra, attention_mask], dim=2)
        if mask_pad is not None:
            if mask_pad.size() != (batch_size, seq_len):
                raise ValueError(f"given mask_pad should be of size {(batch_size, seq_len)}, got {mask_pad.size()}")

        # layer forward
        for i, layer in enumerate(self.layers):
            layer_cache = caches[i] if caches is not None else None
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)
            hidden_states, cache = layer(
                hidden_states,
                attention_mask=attention_mask,
                mask_pad=mask_pad,
                position_ids=position_ids,
                caches=layer_cache,
                use_cache=use_cache,
            )
            if use_cache:
                new_caches = new_caches + (cache,)
        hidden_states = self.layer_norm(hidden_states)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        return tuple([v for v in [hidden_states, all_hidden_states, new_caches] if v is not None])



import torch
import torch.nn as nn

from auv.modules.codec import AUVCodecDec, AUVCodecEncWithVQ


class AUV(nn.Module):
    def __init__(self):
        super().__init__()

    def from_pretrained(self, ckpt_path, device="cuda"):
        ckpt = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        self.tokenizer = AUVCodecEncWithVQ(
            sample_rate=16000,
            enc_conf={
                "hop_length": 320,
                "num_layers": 8,
                "hidden_size": 512,
                "ffn_mult": 4,
                "max_position_embeddings": 4096,
            },
            vq_conf={
                "vq_type": "fvq",
                "num_quantizers": 1,
                "dim": 512,
                "codebook_size": 20480,
                "codebook_dim": 8,
                "commitment": 0.25,
                "codebook_partition_maps": {
                    "speech": [0, 8192],
                    "vocal": [0, 12288],
                    "music": [0, 20480],
                    "audio": [0, 20480],
                    "other": [12288, 20480],
                },
            },
        )
        self.token2wav = AUVCodecDec(
            dec_conf={
                "hop_length": 320,
                "num_layers": 12,
                "hidden_size": 512,
                "ffn_mult": 4,
                "max_position_embeddings": 4096,
            },
        )
        self.load_state_dict(ckpt)
        del ckpt

    def forward(self, feature):
        raise NotImplementedError

    @torch.no_grad()
    def encode(self, data):
        wav_input = data["sample"]
        sample_rate = data["sample_rate"]
        assert wav_input.size(0) == 1, "Only support batch_size == 1 when inference"

        tokenizer_out = self.tokenizer(wav_input, input_sample_rate=sample_rate)
        quantized = tokenizer_out["quantized"]
        tokens = tokenizer_out["tokens"]
        before_quantize = tokenizer_out["before_quantize"]

        res = {
            "quantized": quantized,
            "tokens": tokens,
            "before_quantize": before_quantize,
        }

        return res

    @torch.no_grad()
    def decode(self, emb, **kwargs):
        audio = self.token2wav(emb)
        return audio

    @torch.no_grad()
    def decode_tokens(self, tokens, **kwargs):
        quantized = self.tokenizer.vq2emb(tokens)
        audio = self.token2wav(quantized)
        return audio



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from torch.nn.utils.parametrizations import weight_norm


class FactorizedVectorQuantize(nn.Module):
    def __init__(self, dim, codebook_size, codebook_dim, commitment, **kwargs):
        super().__init__()
        self.codebook_size = codebook_size
        self.codebook_dim = codebook_dim
        self.commitment = commitment

        if dim != self.codebook_dim:
            self.in_proj = weight_norm(nn.Conv1d(dim, self.codebook_dim, kernel_size=1))
            self.out_proj = weight_norm(nn.Conv1d(self.codebook_dim, dim, kernel_size=1))
        else:
            self.in_proj = nn.Identity()
            self.out_proj = nn.Identity()

        self._codebook = nn.Embedding(codebook_size, self.codebook_dim)

    @property
    def codebook(self):
        return self._codebook

    def forward(self, z, codebook_partitions=None):
        """Quantized the input tensor using a fixed codebook and returns
        the corresponding codebook vectors

        Parameters
        ----------
        z: torch.Tensor[B x D x T]
        codebook_partitions: List[List[int]], optional
            List of codebook indices partitions, same for each quantizer
            (ex: [[0, 1024], [1024, 2048]] to allocate for z[0] and z[1])

        Returns
        -------
        z_q: torch.Tensor[B x D x T]
            Quantized continuous representation of input
        vq_loss: Tensor[B]
            Commitment loss to train encoder to predict vectors closer to codebook entries
            plus Codebook loss to update the codebook
        indices: torch.Tensor[B x T]
            Codebook indices (quantized discrete representation of input)
        z_e: torch.Tensor[B x D x T]
            Projected latents (continuous representation of input before quantization)
        """

        # Factorized codes project input into low-dimensional space
        z_e = self.in_proj(z)
        z_q, indices = self.decode_latents(z_e, codebook_partitions=codebook_partitions)

        # Compute commitment loss and codebook loss
        if self.training:
            commitment_loss = F.mse_loss(z_e, z_q.detach(), reduction="none").mean([1, 2]) * self.commitment
            codebook_loss = F.mse_loss(z_q, z_e.detach(), reduction="none").mean([1, 2])
            commit_loss = commitment_loss + codebook_loss
        else:
            commit_loss = torch.zeros(z.shape[0], device=z.device)

        z_q = z_e + (z_q - z_e).detach()

        z_q = self.out_proj(z_q)

        return z_q, indices, commit_loss

    def vq2emb(self, indices, proj=True):
        z_q = self.decode_code(indices)
        if proj:
            z_q = self.out_proj(z_q)
        return z_q

    def get_emb(self):
        return self.codebook.weight

    def embed_code(self, embed_id):
        return F.embedding(embed_id, self.codebook.weight)

    def decode_code(self, embed_id):
        return self.embed_code(embed_id).transpose(1, 2)

    def decode_latents(self, latents, codebook_partitions=None):
        B, D, T = latents.shape
        codebook = self.codebook.weight  # codebook: (N x D)
        encodings = rearrange(latents, "b d t -> (b t) d")

        # L2 normalize encodings and codebook
        encodings = F.normalize(encodings)
        codebook = F.normalize(codebook)

        # Compute euclidean distance with codebook. If l2-normalized, equal to cosine distance
        dist = (
            encodings.pow(2).sum(1, keepdim=True)
            - 2 * encodings @ codebook.t()
            + codebook.pow(2).sum(1, keepdim=True).t()
        )

        if codebook_partitions is not None:
            for i, (start, end) in enumerate(codebook_partitions):
                dist[i * T : (i + 1) * T, :start] = float("inf")
                dist[i * T : (i + 1) * T, end:] = float("inf")

        indices = rearrange((-dist).max(1)[1], "(b t) -> b t", b=B)
        z_q = self.decode_code(indices)

        return z_q, indices



import sys
from typing import Any, Dict, Optional

import torch
import torch.nn as nn
from torchaudio.functional import resample

from auv.modules.conformer import ConformerEncoder
from auv.modules.residual_vq import ResidualVQ
from auv.modules.stft import ISTFTHead, STFTHead


class AUVCodecEncWithVQ(nn.Module):
    def __init__(
        self,
        *,
        enc_conf: Dict[str, Any],
        vq_conf: Optional[Dict[str, Any]] = None,
        init_model: Optional[str] = None,
        sample_rate: int = 24000,
    ):
        super().__init__()

        self.sample_rate = sample_rate
        self.dim = enc_conf["hidden_size"]
        self.hop_length = enc_conf["hop_length"]
        self.stft_head = STFTHead(dim=self.dim, n_fft=self.hop_length * 4, hop_length=self.hop_length)
        self.conformer = ConformerEncoder(**enc_conf)

        # VQ
        if vq_conf is None:
            vq_conf = {
                "vq_type": "fvq",
                "num_quantizers": 1,
                "dim": 512,
                "codebook_size": 8192,
                "codebook_dim": 8,
                "commitment": 0.25,
                "codebook_partition_maps": None,
            }
        self.codebook_partition_maps = vq_conf.pop("codebook_partition_maps")
        if self.codebook_partition_maps is not None:
            assert "all" not in self.codebook_partition_maps
            self.codebook_partition_maps["all"] = [0, sys.maxsize]
        self.vq = ResidualVQ(**vq_conf)

        # init model
        if init_model is not None:
            model_dict = torch.load(init_model, map_location="cpu")
            self.load_state_dict(model_dict)

    def train(self, mode=True):
        super().train(mode)

    def get_hidden_feat(self, x, input_sample_rate=None):
        assert x.dim() == 2, "wav input for auv encoder should with dim==2"
        # resample
        if input_sample_rate is not None and input_sample_rate != self.sample_rate:
            x = resample(x, orig_freq=input_sample_rate, new_freq=self.sample_rate)

        feat = self.stft_head(x)
        (feat,) = self.conformer(feat)
        feat = feat.transpose(1, 2)

        return feat

    def forward(self, wav_input, wav_types=None, input_sample_rate=None, **kwargs):
        hidden_feats = self.get_hidden_feat(wav_input, input_sample_rate=input_sample_rate)

        codebook_partitions = (
            [self.codebook_partition_maps[wav_t] for wav_t in wav_types]
            if self.codebook_partition_maps is not None and wav_types is not None
            else None
        )
        quantized, indices, vq_loss = self.vq(hidden_feats, codebook_partitions=codebook_partitions)

        vq_loss_items = {"vq_loss": vq_loss.sum().item()}
        res = {
            "quantized": quantized.transpose(1, 2),
            "tokens": indices,
            "vq_loss": vq_loss.sum(),
            "vq_loss_items": vq_loss_items,
            "before_quantize": hidden_feats.transpose(1, 2),
        }
        return res

    def vq2emb(self, token):
        x = self.vq.vq2emb(token)
        return x.transpose(1, 2)


class AUVCodecDec(nn.Module):
    def __init__(
        self,
        *,
        dec_conf: Dict[str, Any],
        distill_depth: int = -1,
    ):
        super().__init__()

        self.dim = dec_conf["hidden_size"]
        self.hop_length = dec_conf["hop_length"]
        self.conformer = ConformerEncoder(**dec_conf)
        self.istft_head = ISTFTHead(dim=self.dim, n_fft=self.hop_length * 4, hop_length=self.hop_length)

        self.distill_depth = distill_depth

    def forward(self, x, **kwargs):
        if kwargs.get("return_hidden_feat", False):
            x, all_hidden_feats = self.conformer(x, output_hidden_states=True)
            hidden_feat = all_hidden_feats[self.distill_depth]  # [B, T, C]
        else:
            (x,) = self.conformer(x)

        x = self.istft_head(x).unsqueeze(1)

        if kwargs.get("return_hidden_feat", False):
            return x, hidden_feat

        return x
